/*
 * This program triggers an #BR exception and verifies that the instruction
 * pointer in the trap frame points to the "BOUND" instruction.
 *
 * In other words it makes sure that the #BR exception is a "fault-style"
 * exception.
 *
 * More debug information will be available on the console by setting
 * 'sysctl machdep.uprintf_signal' to '1'.
 */

#ifndef __i386__
#error "BOUND only available in x86 32-bit mode"
#endif

#include <sys/types.h>
#include <sys/sysctl.h>

#include <x86/psl.h>
#include <machine/cpufunc.h>

#include <stdio.h>
#include <signal.h>
#include <assert.h>

#include "bound.h"

void my_bound(void);
void before_bound(void) __asm(__STRING(before_bound));
void after_bound(void) __asm(__STRING(after_bound));

struct bound_array {
	u_int	low;
	u_int	high;
};

extern struct bound_array bound_array;

static int got_fpe;
static int fpe_code;
static void *fpe_addr;

static void
fpe_handler(int num, siginfo_t *info, void *uap)
{
	got_fpe = 1;
	fpe_code = info->si_code;
	fpe_addr = info->si_addr;

	/*
	 * Set the bound_array 'low' and 'high' limits properly so when
	 * 'BOUND' is executed next it will succeed.
	 */
	bound_array.low = BOUND_ARRAY_LOW;
	bound_array.high = BOUND_ARRAY_HIGH;
}

int
main(void)
{
	u_long flags;
	int error;
	struct sigaction sa;

	sa.sa_sigaction = fpe_handler;
	sa.sa_flags = SA_SIGINFO;
	sigemptyset(&sa.sa_mask);
	error = sigaction(SIGFPE, &sa, NULL);
	assert(error == 0);

	my_bound();

	printf("before_bound: %p\n", before_bound);
	printf("after_bound: %p\n", after_bound);
	printf("fpe_addr: %p\n", fpe_addr);
	printf("fpe_code = %d\n", fpe_code);
	printf("got_fpe: %d\n", got_fpe);
	printf("\n");

	if (got_fpe && fpe_code == FPE_FLTSUB && fpe_addr == before_bound)
		printf("success\n");
	else
		printf("failed\n");
}
